part of dartcoin;

class ECKey {
  
  static final ECCurve _CURVE = new fp.ECCurve(
    new BigInteger("fffffffffffffffffffffffffffffffffffffffffffffffffffffffefffffc2f"), //q
    new BigInteger("0000000000000000000000000000000000000000000000000000000000000000"), //a
    new BigInteger("0000000000000000000000000000000000000000000000000000000000000007") //b
  );
  
  static final ECDomainParameters _ECPARAMS = new ECDomainParametersImpl("secp256k1", 
      _CURVE, 
      _CURVE.decodePoint( new BigInteger("0479be667ef9dcbbac55a06295ce870b07029bfcdb2dce28d959f2815b16f81798483ada7726a3c4655da4fbfc0e1108a8fd17b448a68554199c47d08ffb10d4b8").toByteArray() ), //G
      new BigInteger("fffffffffffffffffffffffffffffffebaaedce6af48a03bbfd25e8cd0364141"), //n
      new BigInteger("01") //h
  );
  
  static final BigInteger _HALF_CURVE_ORDER = new BigInteger("fffffffffffffffffffffffffffffffebaaedce6af48a03bbfd25e8cd0364141").shiftRight(1);
  
  static final SecureRandom _secureRandom = new AutoSeedBlockCtrRandom(new AESFastEngine());
  
  BigInteger _priv;
  Uint8List _pub;
  // chache
  Uint8List _pubKeyHash;
  
  /**
   * Create a keypair from a private or public key.
   * 
   * If provided both, they will be considered correctly matching each other.
   * If public key is null, it will be calculated when it is required.
   * If private key is null, this will represent only a public key.
   * If both are null, a new random key will be generated like using ECKey.generate().
   */
  factory ECKey([Uint8List publicKey, BigInteger privateKey]) {
    if(publicKey == null && privateKey == null) return new ECKey.generate();
    return new ECKey._internal(privateKey, publicKey);
  }
  
  /**
   * The private key is retrieved using `new BigInteger.fromBytes(1, privateKey)`.
   */
  factory ECKey.fromBytes(Uint8List privateKey, [Uint8List publicKey, bool compressed = true]) {
    if(privateKey == null) return new ECKey._internal(null, publicKey);
    BigInteger pk = new BigInteger.fromBytes(1, privateKey);
    if(publicKey == null) publicKey = publicKeyFromPrivateKey(pk, compressed);
    return new ECKey._internal(pk, publicKey);
  }
  
  ECKey.fromBigIntegers(BigInteger privateKey, [BigInteger publicKey]) {
    _priv = privateKey;
    if(publicKey != null) {
      _pub = Utils.bigIntegerToBytes(publicKey, 65);
    }
  }
  
  factory ECKey.fromPublic(Uint8List publicKey) => new ECKey._internal(null, publicKey);
  
  /**
   * Intended for internal use only.
   */
  ECKey._internal(BigInteger this._priv, Uint8List this._pub);
  
  /**
   * Generate a new random key pair.
   */
  factory ECKey.generate() {
    //TODO
  }
  
  Uint8List get publicKey => _pub;
  
  Uint8List get pubKeyHash {
    if(_pubKeyHash == null)
      _pubKeyHash = Utils.sha256hash160(_pub);
    return _pubKeyHash;
  }

  /**
   * Returns whether this key is using the compressed form or not. Compressed pubkeys are only 33 bytes, not 64.
   */
  bool get isCompressed {
    return _pub.length == 33;
  }
  
  /**
   * 32-bytes private key
   */
  Uint8List get privateKey {
    return Utils.bigIntegerToBytes(_priv, 32);
  }
  
  bool get hasPrivKey => _priv != null;
  
  bool get isEncrypted {
    //TODO
  }
  
  //TODO maybe this should be moved to Address.fromECKey so that NetworkParams can be taken into account
  Address get address {
    return new Address(Utils.sha256hash160(_pub));
  }
  
  String toString() {
    StringBuffer sb = new StringBuffer()
      ..write("pub:")
      ..write(Utils.bytesToHex(_pub));
    if(isEncrypted) sb.write(" encrypted");
    return sb.toString();
  }
  
  String toStringWithPrivateKey() {
    if(!hasPrivKey) return toString();
    StringBuffer sb = new StringBuffer()
      ..write(toString())
      ..write(" priv:")
      ..write(Utils.bytesToHex(_priv.toByteArray()));
    return sb.toString();
  }
  
  /**
   * This method irreversibly deletes the private key from memory.
   * The key will still be usable as a public key only.
   */
  void clearPrivateKey() {
    _priv = null;
    //TODO encrypted private key
  }
  
  bool operator ==(ECKey other) {
    if(!(other is ECKey)) return false;
    if(identical(this, other)) return true;
    return Utils.equalLists(publicKey, other.publicKey);
  }
  
  int get hashCode {
    return (publicKey[0] & 0xff) | ((publicKey[1] & 0xff) << 8) | ((publicKey[2] & 0xff) << 16) | ((publicKey[3] & 0xff) << 24);
  }
  
  /**
   * Retrieve the public key from the given private key.
   * Use `new BigInteger.fromBytes(signum, magnitude)` to convert a byte array into a BigInteger.
   */
  static Uint8List publicKeyFromPrivateKey(BigInteger privateKey, [bool compressed = true]) {
    ECPoint point = _ECPARAMS.G * privateKey;
    // TODO compression? in java: if(compress) point = compressPoint(point);
    return point.getEncoded();
  }
  
  //TODO implement signing, verification methods
  
  //TODO encryption
  
}